import logging

SERVER_LOGGER = logging.getLogger('server')


class Port:
    """
    Класс - дескриптор для номера порта.
    Позволяет использовать только порты с 1023 по 65536.
    При попытке установить неподходящий номер порта генерирует исключение.
    """
    def __set__(self, instance, value):
        if value not in range(1024, 65536):
            SERVER_LOGGER.critical(
                f'Попытка запуска сервера с указанием неподходящего порта {value}. Допустимы адреса с 1024 до 65535.')
            exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name




def server_help():
    print('Поддерживаемые команды:\n'
          'users - список пользователей\n'
          'connected - список подключённых пользователей\n'
          'history - история входов пользователя\n'
          'exit - завершение работы сервера.\n'
          'help - справка')


def client_help():
    print('Поддерживаемые команды:\n'
          'message - отправить сообщение. Кому и текст будет запрошены отдельно.\n'
          'history - история сообщений\n'
          'contacts - список контактов\n'
          'edit - редактирование списка контактов\n'
          'help - вывести подсказки по командам\n'
          'exit - выход из программы')



import dis


class ClientVerifier(type):
    """
    Метакласс, проверяющий, что в результирующем классе нет серверных
    вызовов таких как: accept, listen. Также проверяется, что сокет не
    создаётся внутри конструктора класса.
    """
    def __init__(cls, cls_name, bases, cls_dict):
        LOAD_GLOBAL = set()
        for func in cls_dict:
            try:
                opcode = dis.get_instructions(cls_dict[func])
            except TypeError:
                pass
            else:
                for _ in opcode:
                    if _.opname == 'LOAD_GLOBAL':
                        LOAD_GLOBAL.add(_.argval)
        for command in ('accept', 'listen', 'socket'):
            if command in LOAD_GLOBAL:
                raise TypeError('В классе обнаружено использование запрещённого метода')
        if 'get_message' in LOAD_GLOBAL or 'send_message' in LOAD_GLOBAL:
            pass
        else:
            raise TypeError('Отсутствуют вызовы функций, работающих с сокетами.')
        super().__init__(cls_name, bases, cls_dict)


class ServerVerifier(type):
    """
    Метакласс, проверяющий, что в результирующем классе нет клиентских
    вызовов таких как: connect. Также проверяется, что серверный
    сокет является TCP и работает по IPv4 протоколу.
    """
    def __init__(cls, cls_name, bases, cls_dict):
        LOAD_GLOBAL = set()
        LOAD_METHOD = set()
        LOAD_ATTR = set()
        for func in cls_dict:
            try:
                opcode = dis.get_instructions(cls_dict[func])
            except TypeError:
                pass
            else:
                for _ in opcode:
                    if _.opname == 'LOAD_GLOBAL':
                        LOAD_GLOBAL.add(_.argval)
                    elif _.opname == 'LOAD_METHOD':
                        LOAD_METHOD.add(_.argval)
                    elif _.opname == 'LOAD_ATTR':
                        LOAD_ATTR.add(_.argval)
        # print(20*'-', 'methods', 20*'-')
        # print(LOAD_GLOBAL)
        # print(20*'-', 'methods_2', 20*'-')
        # print(LOAD_METHOD)
        # print(20*'-', 'attrs', 20*'-')
        # print(LOAD_ATTR)

        if 'connect' in LOAD_GLOBAL:
            raise TypeError('Использование метода connect недопустимо в серверном классе')
        if not ('SOCK_STREAM' in LOAD_ATTR and 'AF_INET' in LOAD_ATTR):
            raise TypeError('Некорректная инициализация сокета.')
        super().__init__(cls_name, bases, cls_dict)



import os
import argparse
import logging
import configparser
import logs.server_log_config

from common.variables import *
from common.utils import *
from common.log_decorator import log
from server_app.core import MessageProcessor
from server_app.server_db import ServerDatabase
from server_app.main_window import MainWindow
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt

# Инициализация логирования сервера.
SERVER_LOGGER = logging.getLogger('server')

if sys.platform == 'linux':
    os.environ['XDG_SESSION_TYPE'] = 'x11'


@log
def arg_parser(default_port, default_address):
    """Парсер аргументов командной строки."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', default=default_port, type=int, nargs='?')
    parser.add_argument('-a', default=default_address, nargs='?')
    parser.add_argument('--no_gui', action='store_true')
    namespace = parser.parse_args(sys.argv[1:])
    listen_address = namespace.a
    listen_port = namespace.p
    gui_flag = namespace.no_gui
    return listen_address, listen_port, gui_flag


@log
def config_load():
    """Парсер конфигурационного ini файла."""
    config = configparser.ConfigParser()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    config.read(f"{dir_path}/{'server.ini'}")
    # Если конфиг файл загружен правильно, запускаемся, иначе конфиг по
    # умолчанию.
    if 'SETTINGS' in config:
        return config
    else:
        config.add_section('SETTINGS')
        config.set('SETTINGS', 'Default_port', str(DEFAULT_PORT))
        config.set('SETTINGS', 'Listen_Address', '')
        config.set('SETTINGS', 'Database_path', '')
        config.set('SETTINGS', 'Database_file', 'server_db.db3')
        return config


@log
def main():
    """Основная функция"""
    # Загрузка файла конфигурации сервера
    config = config_load()
    # Загрузка параметров командной строки, если нет параметров, то задаём
    # значения по умолчанию.
    listen_address, listen_port, gui_flag = arg_parser(config['SETTINGS']['Default_port'],
                                                       config['SETTINGS']['Listen_Address'])
    # Инициализация базы данных
    database = ServerDatabase(os.path.join(config['SETTINGS']['Database_path'],
                                           config['SETTINGS']['Database_file']))
    # Создание экземпляра класса - сервера и его запуск:
    server = MessageProcessor(listen_address, listen_port, database)
    server.daemon = True
    server.start()
    # Если указан параметр без GUI, то запускаем простенький обработчик
    # консольного ввода
    if gui_flag:
        while True:
            command = input('Введите exit для завершения работы сервера.')
            if command == 'exit':
                server.running = False
                server.join()
                break
    # Если не указан запуск без GUI, то запускаем GUI:
    else:
        # Создаём графическое окружение для сервера:
        server_app = QApplication(sys.argv)
        server_app.setAttribute(Qt.AA_DisableWindowContextHelpButton)
        main_window = MainWindow(database, server, config)
        # Запускаем GUI
        server_app.exec_()
        # По закрытию окон останавливаем обработчик сообщений
        server.running = False


if __name__ == '__main__':
    main()
